import java.util.*;
import org.apache.commons.math3.util.Combinations;

import org.apache.commons.lang3.ArrayUtils;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.File;
import java.io.IOException;

import java.sql.Timestamp;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;


/**
 * Item with weight and value.
 * Internal use only.
 */
class KnapsackItem implements Comparable<KnapsackItem> {
	private long weight;
	private long value;

	public KnapsackItem(long weight, long value) {
		this.weight = weight;
		this.value = value;
	}

	public long getWeight() {
		return weight;
	}

	public long getValue() {
		return value;
	}

	public int compareTo(KnapsackItem item) {
		final long score = item.getWeight() * getValue();
		final long itemScore = getWeight() * item.getValue();
		final long diff = score - itemScore;

		// prevent overflow
		if (diff < 0) {
			return -1;
		} else if (diff > 0) {
			return 1;
		} else {
			return 0;
		}
	}
}

/**
 * Instance of the Knapsack Problem.
 * Total weight to be packed: getCapacity()
 * #items: getNumItems()
 * weight/value of i-th item: getWeight(i) / getValue(i)
 * sort() sorts items by (value/weight), in descending order
 */
class KnapsackInstance {
	private long capacity;
	private ArrayList<KnapsackItem> items;

	public KnapsackInstance(int size) {
		capacity = 0;
		items = new ArrayList<KnapsackItem>(size);
	}

	public long getCapacity() {
		return capacity;
	}

	public void setCapacity(long capacity) {
		this.capacity = capacity;
	}

	public void append(long weight, long value) {
		items.add(new KnapsackItem(weight, value));
	}

	public long getWeight(int i) {
		return items.get(i).getWeight();
	}

	public long getValue(int i) {
		return items.get(i).getValue();
	}

	public int getNumItems() {
		return items.size();
	}

	public void sort() {
		Collections.sort(items);
		Collections.reverse(items); // descending order
	}

	public String toString() {
		String description = "capacity=" + getCapacity();

		for (int i = 0; i < getNumItems(); ++i) {
			description += " / (weight_" + i + ", value_" + i + ") = (";
			description += getWeight(i);
			description += ", " + getValue(i) + ")";
		}

		return description;
	}
}

/**
 * This class stores solutions of a Knapsack instance.
 * The i-th boolean indicates whether the i-th item has been chosen.
 * The resulting weight/value/feasibility of the current solution is tracked.
 * Copy instances with copyFrom(...).
 */
class KnapsackSolution {
	KnapsackInstance knapsack;
	double[] choices;
	long weight;
	long value;

	public KnapsackSolution(KnapsackInstance kp) {
		knapsack = kp;
		choices = new double[kp.getNumItems()];
		clear();
	}

	public double get(int i) {
		return choices[i];
	}

	public void set(int i, double b) {
		if (choices[i] != b) {
			weight -= choices[i]*knapsack.getWeight(i);
			value -= choices[i]*knapsack.getValue(i);
			choices[i] = b;
			weight += (long)(b*(double)knapsack.getWeight(i));
			value += (long)(b*(double)knapsack.getValue(i));
		}
	}

	public long getWeight() {
		return weight;
	}

	public long getValue() {
		return value;
	}
	
	
	public int getSize() {
		return choices.length;
	}
	
	public double getCapacity() {
		return knapsack.getCapacity();
	}

	public boolean isFeasible() {
		return getWeight() <= knapsack.getCapacity();
	}

	public void clear() {
		for (int i = 0; i < choices.length; ++i) {
			choices[i] = 0;
		}

		weight = 0;
		value = 0;
	}

	public void copyFrom(KnapsackSolution solution) {
		knapsack = solution.knapsack;
		choices = new double[solution.choices.length];
		for (int i = 0; i < solution.choices.length; ++i) {
			choices[i] = solution.choices[i];
		}
		weight = solution.weight;
		value = solution.value;
	}

	public String toString() {
		String description
				= (isFeasible() ? "" : "INFEASIBLE! ")
				+ "Total Weight: " + getWeight()
				+ " / total value: " + getValue() + " / items:";
		for (int i = 0; i < choices.length; ++i) {
			if (choices[i] == 1) {
				description += " " + i;
			}else if (choices[i] > 0){
				description += " (" + String.format("%.2f",choices[i]) + ")*" + i;
			}
		}

		return description;
	}
}

/**
 * Here are the actual algorithms!
 */
public class Knapsack1234567 {
	// Parses input. Do not edit.
	public static KnapsackInstance readInstance(String filePath) {
		try{
		File file = new File(filePath);
		Scanner scanner = new Scanner(file);
		KnapsackInstance kp = new KnapsackInstance(scanner.nextInt());

		while (scanner.hasNextLong()) {
			long capacity = scanner.nextLong();

			if (scanner.hasNextLong()) {
				long weight = scanner.nextLong();
				long value = scanner.nextLong();
				kp.append(weight, value);
			} else {
				kp.setCapacity(capacity);
			}
		}

		kp.sort();
		return kp;
		}catch(Exception e){}
		return null;
	}
	
	public static void writeToCSV(Algorithmus alg, String name, KnapsackSolution ks, long time_init, long time_solve, int k){
		String file_name = alg + "";
		if (alg == Algorithmus.APPROXIMATION){
			file_name += k+"";
		}
		file_name += ".csv";
		System.out.println("value: " + (int)ks.getValue());
		System.out.println(time_init+"ms");
		System.out.println(time_solve+"ms");
		
		try{
		Path filePath = Paths.get(file_name);
		if (!Files.exists(filePath)) {
    		Files.createFile(filePath);
			Files.write(filePath, "Datei Name;Itemzahl;Kapazität;Wert;Zeit (Einlesen);Zeit (Lösen)\n".getBytes(), StandardOpenOption.APPEND);
		}
		String s = name + ";"+ ks.getSize()+ ";"+ ks.getCapacity() +";" + (int)(ks.getValue()) + ";" + time_init + ";" + time_solve + "\n";
		Files.write(filePath, s.getBytes(), StandardOpenOption.APPEND);
		}catch(Exception e){}
	}

	public static KnapsackSolution bruteForce(KnapsackInstance kp) {
		//TODO
		return null; 
	}
	
	public static KnapsackSolution fractional(KnapsackInstance kp) {
		//TODO
		return null; 
	}
	
	public static KnapsackSolution greedy(KnapsackInstance kp) {
		//TODO
		return null; 
	}

	public static KnapsackSolution dynamicProgramming(KnapsackInstance kp) {
		//TODO
		return null; 
	}
	
	public static KnapsackSolution branchAndBound(KnapsackInstance kp) {
		//TODO
		return null; 
	}


	public static KnapsackSolution greedyK(KnapsackInstance kp, int k) {
		//TODO
		return null; 
	}
	
	public enum Algorithmus{
		GREEDY, FRACTIONAL, BRANCH, DYNAMIC, BRUTEFORCE, APPROXIMATION	
	}
	
public static void main(String[] args) {		
		//Wähle den Algorithmus hier:
		Algorithmus alg = Algorithmus.APPROXIMATION;
		int k=0;
		try{
		 k = Integer.parseInt(args[1]); //Dieser Parameter ist für Greedy_k
		}catch(Exception e){}
		long time_init, time_solve;
		Timestamp time_start = new Timestamp(System.currentTimeMillis());
		KnapsackInstance kp = readInstance(args[0]);

		Timestamp time_end = new Timestamp(System.currentTimeMillis());
		time_init = time_end.getTime()-time_start.getTime();
		//System.out.println(kp);
		KnapsackSolution ks = new KnapsackSolution(kp);
		time_start = new Timestamp(System.currentTimeMillis());
		
		switch (alg){
			case FRACTIONAL:
				ks.copyFrom(fractional(kp));
				break;
			case GREEDY: 
				ks.copyFrom(greedy(kp));
				break;
			case BRUTEFORCE:
				ks.copyFrom(bruteForce(kp));
				break;
			case BRANCH:
				ks.copyFrom(branchAndBound(kp));
				break;
			case DYNAMIC:
				ks.copyFrom(dynamicProgramming(kp));
				break;
			case APPROXIMATION:
				ks.copyFrom(greedyK(kp, k));
				break;
			default:
				System.out.println("Opps...something went wrong. Correct enum chosen?");
				return;
				
		}
		time_end = new Timestamp(System.currentTimeMillis());
		time_solve = time_end.getTime()-time_start.getTime();
//		System.out.println(ks);
		writeToCSV(alg, args[0], ks, time_init, time_solve,k);
		
	}
}
